#ifndef RigidCLLH
#define RigidCLLH

#include <vector>
#include <list>
#include <string>

/** \file RigidCLL.h

    \brief Headers of the RigidCLL implementation.

    Headers used in the implementation of the RigidCLL algorihtm. This
    algorithm is the one described in

    - V. Ruiz de Angulo, J. Cort&eacute;s, J. M. Porta, Rigid-CLL: Avoiding Constant-distance Computations in Cell Linked-Lists Algorithms, Journal of Computational Chemistry, Vol. 33, No. 3, pp. 294-300, 2012
    .
*/

using namespace std;

/** 
    \brief Debug flag.
      
    If set to 1 some extra checks are performed. 
    Set it to 0 to speed up the execution.
    Set to 2 to visualize more information during the execution.
*/
#ifndef RIGIDCLL_DEBUG
  #define RIGIDCLL_DEBUG 1
#endif

/** 
    \brief Homogeneous transform.

    4x4 matrix that encodes an homogeneous transform.
*/
typedef double RigidTransform[4][4];
 
/** 
    \brief Atom information.

    Information stored for each atom in the system.
*/
class AtomInfo {
 public:
  unsigned int id;      /**< Atom identifier, as provided by the caller */
  unsigned int rID;     /**< Identifier of the rigid including the atom. */
  double op[3];         /**< Coordinates for the atom original postion. */
  double cp[3];         /**< Coordinates for the atom current postion. */
  double r;             /**< Atom radius */
  unsigned int c[3];    /**< Indexes for the cell including the atom. */

  
  /** 
      \brief Constructor.

      AtomInfo constructor.

      \param identifier External identifier for this atom.
      \param rigid Tbe identifier of the rigid group including this atom.
      \param x X coordinate for the atom.
      \param y Y coordinate for the atom.
      \param z Z coordinate for the atom.
      \param radius Radius for the atom.
  */
  AtomInfo(unsigned int identifier,unsigned int rigid,double x,double y,double z,double radius);

  /** 
      \brief Moves an atom.

      Moves an atom using the provided rigid transform.

      \param t The transform to use in the motion.
  */
  inline void move(RigidTransform t); 

  /** 
      \brief Distance to another atom.

      Computes the distance to another atom.

      \param a The other atoms.

      \return The distance between the two atoms.
  */
  double distance(AtomInfo *a);
};

/** 
    \brief Prints the atom info.
      
    Prints the information stored in the atom.
*/
ostream& operator<<(ostream& os,const AtomInfo &a);

/** 
    \brief Pair of atoms identifiers.

    Pairs of atom identifiers returned when detecting nearby atoms.
*/
class AtomPair {
 public:
  unsigned int a1;    /**< Identifier of the first atom (as given by the caller when inserting the atom). */
  unsigned int r1;    /**< Identifier for rigid group including the first atom. */
  unsigned int a2;    /**< Identifier of the second atom (as given by the caller when inserting the atom). */
  unsigned int r2;    /**< Identifier for rigid group including the second atom. */
  double d2;          /**< Squared distance between the atoms */

  /** 
      \brief Constructor
      
      Defines a new AtomPair.

     \param atom1 The information of the first atom.
     \param atom2 The information of the second atom.
     \param dst2 Squared distance between the two atoms.
  */
  AtomPair(AtomInfo *atom1,AtomInfo *atom2,double dst2);

  /** 
      \brief Function to sort atom pairs.
    
      Function to sort the output of the interaction detection. 
       
      \param a The other atom pair.
       
      \return true if the current AtomPair is lower than 'a'
  */
  bool operator<(const AtomPair a) const;
  
  /** 
      \brief Function to compare atom pairs.
    
      Function to compare the output of the interaction detection. 
      This is only used when comparing the output of the algorihtms.
       
      \param a The other atom pair.
       
      \return true if the current AtomPair is the same as 'a'

  */
  bool operator==(AtomPair &a);
};

/** 
    \brief Prints the atom pair.
      
    Prints the information stored in the atom pair.
*/
ostream& operator<<(ostream& os,const AtomPair &a);

/** 
    \brief Rigid information.

    Information stored for each rigid group of atoms.
*/
typedef struct {
  unsigned int firstAtom; /**< Index of the first atom in the rigid. */
  unsigned int lastAtom;  /**< Index of the first atom of the next rigid, if any. */
} RigidInfo;

/** 
    \brief Cell information.

    Information stored in each of the cells of the grid (in the actually used ones).
*/
class CellInfo {
 public:
  unsigned int mobileID;             /**< Identifier of the first mobile atom in the cell. */
  unsigned int fixedID;              /**< Identifier of the first fixed atom in the cell. */
  
  /** 
      \brief Constructor.

      CellInfo constructor.
  */
  inline CellInfo();
  /** 
      \brief Reset the cellInfo.

      Resets the information stored into a CellInfo.
      This implies to "forget" about the mobile atoms but not about
      the fixed ones.
  */
  inline void Reset();

  /** 
      \brief Adds a fixed atom to a cell.

      Adds a fixed atom into a cell.

      \param id The identifier of the atom.
      \param fixedLL The linked list with the fixed atoms.
  */
  void AddFixedAtom(unsigned int id,unsigned int *fixedLL);

  /** 
      \brief Adds a mobile atom to a cell.

      Adds a mobile atom into a cell.

      \param id The identifier of the atom.
      \param mobileLL The linked list with the mobile atoms.
  */
  inline void AddMobileAtom(unsigned int id,unsigned int *mobileLL);

};

/** 
    \brief The main RigidCLL class.

    Class encompassing the information and methods implementing the RigidCLL algorithm.
*/
class RigidCLL {

 private:

  vector<AtomInfo> fixedAtoms;         /**< Set of fixed atoms. The position of these atoms never changes. */
  vector<AtomInfo> mobileAtoms;        /**< Set of mobile atoms. */
  vector<RigidInfo> rigids;            /**< Rigid groups of mobile atoms. */
  unsigned int* fixedLL;               /**< Linked list with the fixed atoms. */
  unsigned int* mobileLL;              /**< Linked list with the mobile atoms. */
  CellInfo ***grid;                    /**< Grid representing the environment where atoms can move. */
  double o[3];                         /**< Coordinates giving the origin of the grid. */
  unsigned int s[3];                   /**< Number of cells along each dimension. */
  double cutoff;                       /**< Cutoff value as given by the user. */
  double d;                            /**< Size of the cells (the same in X,Y, and Z). */
  double d2;                           /**< Squared 'd'. */
  unsigned int rID;                    /**< Identifier of the rigid being defined. */
  RigidInfo cr;                        /**< Current rigid. Used during rigid definition. */
  unsigned int nFixed;                 /**< Number of fixed atoms. */
  unsigned int nMobile;                /**< Number of mobile atoms. */
  unsigned int lastInGrid;             /**< Last mobile atom inserted in the grid in the previous query, if any */
  bool **checkInteraction;             /**< Array with the possible interactions between rigids. */


  /** 
      \brief Basic initialization of a RigidCLL object.

      Function shared by all RigidCLL constructors.
  */
  void InitRigidCLL();

  /** 
      \brief Updates the cell assign.

      Update the cell information stored in the AtomInfo giving the indexes
      of the cell including the atom.

      This functions has no effect in the cell, but in the atom info.

      \param a The atom information ot update.
  */
  inline void UpdateCell(AtomInfo *a);

  /** 
      \brief Inserts a fixed atom in  the grid.

      Inserts a fixed atom in the grid at the coordinates already stored
      in the atoms info (possibly using \ref UpdateCell).

      \param n Internal indentifier of the atom.
  */
  void InsertFixedInGrid(unsigned int n);

  /** 
      \brief Inserts a mobile atom in  the grid.

      Inserts a mobile atom in the grid at the coordinates already stored
      in the atoms info (possibly using \ref UpdateCell).

      \param n Internal indentifier of the atom.
  */
  inline void InsertMobileInGrid(unsigned int n);
    

  /** 
      \brief Deallocates the memory used for the grid.

      Release the memory used for the grid. This is done when deleting
      the RigidCLL object and also when calling the \ref SetCutOff method
      when the system is already initialized (i.e., before re-initializing
      the collision detection).
  */
  void DeleteGrid();

  /** 
      \brief Deallocates the interaction table.

      Deallocates the table encoding the possible interactions between rigids.
  */
  void DeleteInteractions();

  /** 
      \brief Allocates the interaction table.

      Allocates the table encoding the possible interactions between rigids and
      initializes it allowing the interaction between all pairs of rigid clusters.

      This is automatically called the first call to \ref ActivateInteraction or
      to \ref DeactivateInteraction. From this point no new rigid can be
      defined.
  */

  void InitInteractions();

  /** 
      \brief Resets the grid from a previous query.

      Remove the mobile atoms inserted in the grid in a previous query.
  */
  inline void ResetQuery();

  /** 
      \brief Moves the atoms.

      Moves the all the mobile atoms.

      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.
  */
  inline void Move(RigidTransform *t);

  /** 
      \brief Steric clash check around a given atoms.

      Checks for an steric clash in the 27 cells surounding a given atom.

      We assume that the cell for the atom is pre-computed before calling
      this method (using \ref UpdateCell).

      In this method we take into account the decomposition in rigids and
      only consider steric clashes between atoms in different rigids.

      This method is shared by \ref StericClash and \ref CLLStericClash.

      \param a1 The atom.

      \return true if a steric clash is detected.
  */
  inline bool Check27(AtomInfo *a1);

 public:

  /** 
      \brief RigidCLL constructor.

      The RigidCLL constructor.
  */
  RigidCLL();

  /** 
      \brief RigidCLL constructor.

      The RigidCLL constructor. Read information from a file to 
      define a RigidCLL structure.

      \param name The file name from where to read the information.
  */
  RigidCLL(char *name);
    
  /** 
      \brief RigidCLL destructor.

      The RigidCLL destructor.
  */
  ~RigidCLL();

  /** 
      \brief Returns the number of rigids.

      Returns the number of rigids in the RigidCLL structure. Note that
      the fixed atoms are counted as one rigid cluster. Thus, the output
      of this function is the number of mobile rigid groups plus one.

      If there is a rigid under construction, it is considered as
      compleately defined (altthogh some atoms can be defined latter on
      to it).
  */
  unsigned int GetNumRigids();


  /** 
      \brief Sets the cutoff.

      Sets the cutoff. Use to to detect steric clashes (this is
      the default value).

      When using this function the internal grid is deleted (if any)
      and it will be re-defined the next call to an interaction
      detection method.
      
      \param c The new cutoff to set.
  */
  void SetCutOff(double c);

  /** 
      \brief Returns the cutoff currently used.

      Returns the cutoff being used to detect atom interactions.
      This is basically the parameter given when calling \ref SetCutOff.

      When aiming for steric clash detection, the returned cutoff is 0.

      \return The current cutoff.
  */
  double GetCutOff();

  /** 
      \brief Size of the bounding box.

      Size of the area coveed by the grid.

      \param axis 0 for X, 1 for Y, 2 for Z.
  */
  double GetBBSize(unsigned int axis);
    
  /** 
      \brief Defines a fixed atom.

      Defines a new fixed atom.
      
      \param id The atom identifier.
      \param x The X coordinate of the atom.
      \param y The Y coordinate of the atom.
      \param z The Z coordinate of the atom.
      \param r The atom radius.
  */
  void AddFixedAtom(unsigned int id,double x,double y,double z,double r);

  /** 
      \brief Defines a fixed atom.

      Defines a new fixed atom.

      \param id The atom identifier.
      \param p The three coordinates defining the atom position.
      \param r The atom radius.
  */
  void AddFixedAtom(unsigned int id,double *p,double r);

  /** 
      \brief Rigid definition.

      Starts the definition of a rigid group of atoms.

      \returns The identifier assigned to the new rigid. Identifiers
               start at 1 (0 is reserved for the 'fixed' rigid).
  */
  unsigned int NewRigid();

  /** 
      \brief Defines an atom.

      Defines a new atom in the current rigid group of atoms.
      
      \param id The atom identifier.
      \param x The X coordinate of the atom.
      \param y The Y coordinate of the atom.
      \param z The Z coordinate of the atom.
      \param r The atom radius.
  */
  void AddAtom(unsigned int id,double x,double y,double z,double r);

  /** 
      \brief Defines an atom.

      Defines a new atom in the current rigid group of atoms.

      \param id The atom identifier.
      \param p The three coordinates defining the atom position.
      \param r The atom radius.
  */
  void AddAtom(unsigned int id,double *p,double r);

  /** 
      \brief Concludes the rigid definition.

      Concludes the definition of a rigid group of atoms.
  */
  void EndRigid();
  
  /** 
      \brief Activates the interaction between a pair of rigids.

      Activates the interaction between a pair of rigids. 
      
      The interactions are set the first time we call \ref ActivateInteraction
      or \ref DeactivateInteraction. If not set, the interactions are assumed
      to be all active.

      When we start giving allowed or disabled intereactions, we can not
      add more rigids until we call the \ref Reset method.

      Note that rigids are numbered from 1 in the order they are defined
      by the user. Rigid 0 refers to the 'fixed' group of atoms.

      \param r1 The identifier for the first rigid group. 0 for the fixed cluster.
      \param r2 The identifier for the second rigid group. 0 for the fixed cluster.
  */
  void ActivateInteraction(unsigned int r1,unsigned int r2);

  /** 
      \brief Dectivates the interaction between a pair of rigids.

      The interactions are set the first time we call \ref ActivateInteraction
      or \ref DeactivateInteraction. If not set, the interactions are assumed
      to be all active.

      When we start giving allowed or disabled intereactions, we can not
      add more rigids until we call the \ref Reset method.
      
      Note that rigids are numbered from 1 in the order they are defined
      by the user. Rigid 0 refers to the 'fixed' group of atoms.

      \param r1 The identifier for the first rigid group. 0 for the fixed cluster.
      \param r2 The identifier for the second rigid group. 0 for the fixed cluster.
  */
  void DeactivateInteraction(unsigned int r1,unsigned int r2);

  /** 
      \brief Initializes the collision detection.

      Intializes the structures required for collision detection.

      Once the WarmUp is used, no more rigids or fixed atoms can be added
      until we use the \ref Reset method.

      In general there is no need to explicitly call this method, as, if
      necessary it will be automatically invoked when first checking for interactions.

      \param t Set of rigid transforms providing a valid pose for the rigid groups.
  */
  void WarmUp(RigidTransform *t);

  /** 
     \brief Reset the internal information.

     Resets the internal information so that new rigids can be added
     to the problem.
     
     Note that both the grid and the interaction table will be deleted.
     Thus, the interaction information must be re-entered before checking
     for interaction (or it will be assumed that all rigid clusters can
     interact between them).
  */
  void Reset();

  /** 
      \brief Checks for steric clashes.
	
      Determines if there is at least one steric clash in the problem. This is the 
      method to be used when the \ref WarmUp is called with cutoff 0.

      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.

      \return true if there is a steric clash.
  */
  bool StericClash(RigidTransform *t);

  /** 
      \brief \ref StericClash using the orignal CLL algorithm.

      \ref StericClash using the orignal CLL algorithm (i.e., Algorithm 2 in the paper).
      This is used to show the advantage of the RigidCLL algorithm.

      Note that this methods takes into account the decomposition of the
      problem in rigid clusters and, thus, it only takes into account
      clashes between atoms in different clusters. This is the same
      as \ref StericClash.
	
      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.

      \return true if there is a steric clash.
  */
  bool CLLStericClash(RigidTransform *t);

  /** 
      \brief A brute force alternative to \ref StericClash.

      A brute force alternative to \ref StericClash used for test and debug.
	
      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.

      \return true if there is a steric clash.
  */
  bool BruteForceStericClash(RigidTransform *t);

    
  /** 
      \brief Detects nearby atoms.
	
      Detects nearby atoms (atoms closer than
      the cutoff). This is the method to sue when 
      the \ref WarmUp is called with cutoff larger than 0.

      Note that only pair of atoms in different rigid groups are considered. This is
      the essence of the RigidCLL algorihtm.

      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.
      \param c Output list with the paris of atoms closer than the cutoff distance.
  */
  void Interactions(RigidTransform *t,list<AtomPair> *c);

  /** 
      \brief The same as \ref Interactions, but using the standard CLL algoritm.

      The same as \ref Interactions but using the standard CLL algoritm. This is used to
      show the advantage of the RigidCLL algorithm.
      
      Note that this methods does not take into account the decomposition of the
      problem in rigid clusters and, thus, it can return interactions between
      atoms in the same rigid (in contrast, \ref Interactions only considers
      interactions between atoms in different clusters). Thus, the list
      of interactions returned by this method is typically larger than the one
      returned by \ref Interactions

      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.
      \param c Output list with the paris of atoms closer than the cutoff distance.
  */
  void CLLInteractions(RigidTransform *t,list<AtomPair> *c);

  /** 
      \brief A brute force alternative to \ref Interactions.

      A brute force alternative to \ref Interactions used for test and debug.

      This method follows the same policy as \ref Interactions (and not the one
      of \ref CLLInteractions): it only returns interactions between atoms
      in different rigid clusters. Thus, the output of this method can be used
      to check the output of \ref Interactions but not the one of \ref CLLInteractions.

      \param t Array of transforms to apply to the rigid groups. Numbered from one. The
               transform t[0] is reserved for the fixed atoms although it is not used in 
               practice. So, t[1] should refer to the transform for the first rigid
               group of atoms.
      \param c Output list with the paris of atoms closer than the cutoff distance.
  */
  void BruteForceInteractions(RigidTransform *t,list<AtomPair> *c);
};

#endif
